{-# LANGUAGE GADTs, ScopedTypeVariables, FlexibleContexts, RankNTypes, TypeFamilies #-}
{-# OPTIONS_GHC -Wall -fno-warn-name-shadowing #-}
module Lam.UI where

import Prelude hiding (log)
import Control.Monad
import Data.IORef
import Data.List
import Data.Map (Map)
import qualified Data.Map as Map

import Data.Cursor.CLASE.Language

import Graphics.UI.Gtk.Glade
import Graphics.UI.Gtk hiding (Cursor)
import qualified Graphics.UI.Gtk.ModelView as New

import Lam.Lam
import Lam.Language
import Lam.Render
import Lam.Binding
import Data.Maybe

data CursorHolder where
  CH :: Cursor Lam a a -> CursorHolder

data GuiState 
  = GS { cursorBuffer :: TextBuffer
       , variableNameEntry :: Entry
       , variableNameDialog :: Dialog
       , whichVariableModel :: New.ListStore (Integer, String)
       , whichVariableDialog :: Dialog
       , cursorH :: IORef CursorHolder
       }


initUI :: IO ()
initUI = do
  initGUI

  Just xml <- xmlNew "Lam/language.glade"

  cursorTextView <- xmlGetWidget xml castToTextView "cursor"
  cursorBuffer <- textViewGetBuffer cursorTextView
  tagTable <- textBufferGetTagTable cursorBuffer
  textTag<- textTagNew (Just $ "default")
  textTag `set` [textTagScale := 3]
  tagTable `textTagTableAdd` textTag


  variableNameEntry <- xmlGetWidget xml castToEntry "variableChoiceEntry"
  variableNameDialog <- xmlGetWidget xml castToDialog "variableChoiceDialog" 
  variableNameDialog `onResponse` (\_ -> widgetHide variableNameDialog)

  variableNameEntry `onEntryActivate` (variableNameDialog 
                                        `dialogResponse` ResponseOk)

  whichVariableTreeView <- xmlGetWidget xml New.castToTreeView "whichVariableTreeView"
  whichVariableModel <- New.listStoreNew []
  whichVariableTreeView `New.treeViewSetModel` whichVariableModel

  whichVariableTreeViewRenderer <- cellRendererTextNew
  
  whichVariableTreeViewColumn <- treeViewColumnNew
  New.cellLayoutPackStart whichVariableTreeViewColumn whichVariableTreeViewRenderer True
  New.cellLayoutSetAttributes whichVariableTreeViewColumn 
                              whichVariableTreeViewRenderer 
                              whichVariableModel 
                              (\(i,s) -> [New.cellText := s ++ "  " ++ show i])
  treeViewAppendColumn whichVariableTreeView whichVariableTreeViewColumn

  whichVariableDialog <- xmlGetWidget xml castToDialog "whichVariableDialog"
  whichVariableDialog `onResponse` (\_ -> widgetHide whichVariableDialog)

  whichVariableTreeView `onRowActivated` 
    (\[idx] _ -> whichVariableDialog `dialogResponse` (ResponseUser (succ idx)))
  
  ref <- newIORef $ CH emptyCursor

  let guiState = GS { 
           cursorBuffer = cursorBuffer
         , variableNameEntry = variableNameEntry
         , variableNameDialog = variableNameDialog
         , whichVariableModel = whichVariableModel
         , whichVariableDialog = whichVariableDialog
         , cursorH = ref
         }

  cursorTextView `onKeyPress` (cursorKeyPress guiState)
  refreshAll guiState

  window <- xmlGetWidget xml castToWindow "mainWindow"
  window `onDestroy` mainQuit
  widgetShowAll window
  
  mainGUI
  
emptyCursor :: Cursor Lam Lam Lam
emptyCursor = Cursor {
  it  = Lam NoExp,
  ctx = Stop,
  log = Route Stop Stop
}


getVariableNameFromDialog :: GuiState -> String -> IO (Maybe String)
getVariableNameFromDialog gs inp = do
  variableNameEntry gs `entrySetText` inp
  respId <- dialogRun $ variableNameDialog gs

  case respId of
    ResponseOk -> do
      respText <- (entryGetText $ variableNameEntry gs)
      return $ Just respText
    _ -> return Nothing
  
getVariableChoiceFromDialog :: GuiState -> Map Integer String -> 
                                           IO (Maybe Integer)
getVariableChoiceFromDialog gs vars = do
  New.listStoreClear (whichVariableModel gs) 
  forM varList (New.listStoreAppend (whichVariableModel gs))
  respId <- dialogRun $ whichVariableDialog gs

  case respId of
    ResponseUser n -> do
      let selectedVar = fst $ varList !! (pred n) 
      return $ Just selectedVar
    _ -> return Nothing
  where
    varList = Map.toList vars
    
cursorKeyPress :: GuiState -> Event -> IO Bool
cursorKeyPress gs (Key { eventKeyChar  = Just char, 
                          eventModifier = modifiers })
  | char == 'h'               = moveCursor [gml, gmu]
  | char == 'j'               = moveCursor [gmd, gmr]
  | char == 'k'               = moveCursor [gmu, gml]
  | char == 'l'               = moveCursor [gmr, gmd]
  | char == '\\'              = modifyCursor insertLambda ((const . const) Nothing)
  | char == ' '               = modifyCursor ((Just . ) . insertAppOrArr (Shift `elem` modifiers)) 
                                             ((Just . ) . moveIntoAppOrArr (Shift `elem` modifiers))
  | char == 't'               = modifyCursor insertTau ((const . const) Nothing)
  | char == 'x'               = modifyCursor ((Just .) . deleteExpOrType) ((const . const) Nothing)
  where
    ref = cursorH gs

    gmd = fmap (\(CWM c _) -> ExistsR c) . genericMoveDown
    gmu = fmap (\(CWM c _) -> ExistsR c) . genericMoveUp
    gml = genericMoveLeft
    gmr = genericMoveRight

    moveCursor :: (forall x a . [Cursor Lam x a -> Maybe (ExistsR Lam (Cursor Lam x))]) -> IO Bool
    moveCursor movs = do
      CH theCursor @ Cursor {} <- readIORef ref

      maybe (return True) 
            (\(ExistsR cursor') -> do
              writeIORef ref $ CH (resetLog cursor')
              refreshAll gs
              return True
            )
            (msum $ map ($ theCursor) movs)

    modifyCursor :: (forall a . TypeRepI a -> a -> Maybe a) -> 
                    (forall a . TypeRepI a -> a -> Maybe (ExistsR Lam (Route Lam a)))  -> IO Bool
    modifyCursor editIt moveFocus = do
      CH theCursor@Cursor{} <- readIORef ref
      maybe (return True)
            (\it' -> do
              let cursor' = Cursor { it = it'
                                   , ctx = ctx theCursor
                                   , log = log theCursor }
              
              mRoute <- return $ moveFocus (unWrap . reify $ undefined) it'

              let outCursorHolder = fromMaybe (CH cursor') $ do
                                      (ExistsR route) <- mRoute
                                      cursor'' <- followRoute cursor' route
                                      return $ CH (resetLog cursor'')

              writeIORef ref outCursorHolder
              refreshAll gs
              return True
            )
            (editIt (unWrap . reify $ undefined) (it theCursor))


{- edit a primitive -}
cursorKeyPress gs (Key { eventKeyName = "Return" }) = do
  CH theCursor @ Cursor {} <- readIORef ref

  mNewValue <- runDialog gs theCursor (unWrap $ reify undefined) (it theCursor)

  maybe (return True)
        (\it' -> do 
          let cursor' = Cursor { it = it'
                               , ctx = ctx theCursor
                               , log = log theCursor
                               }
          let outCH = CH cursor' 
          writeIORef ref outCH
          refreshAll gs
          return True)
        mNewValue
  where
    ref = cursorH gs

    runDialog :: GuiState -> Cursor Lam a a -> TypeRepI a -> a -> IO (Maybe a)
    runDialog gs _ ExpT (Abs str t e) = do
      mNewName <- getVariableNameFromDialog gs str
      return $ fmap (\nstr -> Abs nstr t e) mNewName
    runDialog gs theCursor ExpT NoExp
      | not $ Map.null inScopeVars = do
          mVarChoice <- getVariableChoiceFromDialog gs inScopeVars
          return $ fmap Var mVarChoice
      where
        inScopeVars = inScopeVariables theCursor
    runDialog gs cursor ExpT (Var _)       = runDialog gs cursor ExpT NoExp
    runDialog _ _ _ _ = return Nothing


cursorKeyPress _ _ = return False

insertLambda :: TypeRepI a -> a -> Maybe a
insertLambda ExpT  e        = Just $ (Abs "?" NoType (shiftBindings e))
insertLambda LamT  (Lam e)  = Just $ Lam (Abs "?" NoType (shiftBindings e))
insertLambda TypeT _        = Nothing

insertAppOrArr :: Bool -> TypeRepI a -> a -> a
insertAppOrArr True   ExpT e       = App NoExp e
insertAppOrArr False  ExpT e       = App e NoExp
insertAppOrArr True   LamT (Lam e) = Lam (App NoExp e)
insertAppOrArr False  LamT (Lam e) = Lam (App e NoExp)
insertAppOrArr True   TypeT t      = Arr NoType t
insertAppOrArr False  TypeT t      = Arr t NoType

moveIntoAppOrArr :: Bool -> TypeRepI a -> a -> ExistsR Lam (Route Lam a)
moveIntoAppOrArr True  ExpT  _ = ExistsR (Route Stop (MW MAppToExp0 `Step` Stop))
moveIntoAppOrArr False ExpT  _ = ExistsR (Route Stop (MW MAppToExp1 `Step` Stop))
moveIntoAppOrArr True  LamT  _ = ExistsR (Route Stop (MW MLamToExp `Step` MW MAppToExp0 `Step` Stop))
moveIntoAppOrArr False LamT  _ = ExistsR (Route Stop (MW MLamToExp `Step` MW MAppToExp1 `Step` Stop))
moveIntoAppOrArr True  TypeT _ = ExistsR (Route Stop (MW MArrToType0 `Step` Stop))
moveIntoAppOrArr False TypeT _ = ExistsR (Route Stop (MW MArrToType1 `Step` Stop))

insertTau :: TypeRepI a -> a -> Maybe a
insertTau TypeT NoType = Just Unit
insertTau _ _          = Nothing

deleteExpOrType :: TypeRepI a -> a -> a
deleteExpOrType LamT _  = Lam NoExp
deleteExpOrType ExpT _  = NoExp
deleteExpOrType TypeT _ = NoType

refreshAll :: GuiState -> IO ()
refreshAll gs = do
  CH cursor@Cursor {} <- readIORef (cursorH gs)
  let cursorText = render cursor
  (cursorBuffer gs) `textBufferSetText` cursorText
  join $ liftM2 (textBufferApplyTagByName (cursorBuffer gs) "default") 
                (textBufferGetStartIter $ cursorBuffer gs) 
                (textBufferGetEndIter $ cursorBuffer gs)

unWrap :: TypeRep Lam a -> TypeRepI a
unWrap (TW x) = x

shiftBindings :: Exp -> Exp
shiftBindings = shiftBindings' 0
  where
    shiftBindings' :: Integer -> Exp -> Exp
    shiftBindings' n (Var i)
      | i < n = Var i
      | otherwise = Var (i+1)
    shiftBindings' n (App e1 e2) = App (shiftBindings' n e1) (shiftBindings' n e2)
    shiftBindings' n (Abs str typ e) = Abs str typ (shiftBindings' (succ n) e)
    shiftBindings' _ NoExp = NoExp
